@@ -84,8 +84,10 @@ sub compile {
         #print "match: ", Dumper( Pugs::Grammar::Rule->rule( $self->{source} ) );
         my $ast = Pugs::Grammar::Rule->rule( 
             $self->{source} )->();
-            
-        #$self->{ast} = $ast;
+        
+        # save the ast for debugging
+        $self->{ast} = $ast;
+
         #print "ast: ",Dumper($ast),"\n";
         #die "Error in rule: '$rule_source' at: '$ast->tail'\n" if $ast->tail;
         #print 'rule ast: ', do{use Data::Dumper; Dumper($ast{capture})};
@@ -212,8 +214,10 @@ sub reinstall {
   my $rule = index($name, '::') > -1 ? $name : scalar(caller)."::$name";
   my $slot = qualify_to_ref($rule);
 
-  no warnings 'redefine';
-  *$slot = $class->compile(@etc)->code;
+  eval {
+      no warnings 'redefine';
+      *$slot = $class->compile(@etc)->code;
+  }; warn $@ if $@;
 }
 
 sub install {
